

class LikelihoodScanPlot(object):
    def __init__(self):
        self._data = None
        
    @property
    def data(self):
        return self._data
    
    def load(self, input_files):
        pass
             